from __future__ import annotations

from typing import Hashable

import pandas as pd
from pandas.api.types import is_dict_like
from pandas.api.types import is_number

from dtoolkit._typing import Number
from dtoolkit._typing import SeriesOrFrame
from dtoolkit.accessor.register import register_dataframe_method


@register_dataframe_method
def weighted_mean(
    df: pd.DataFrame,
    /,
    weights: list[Number] | dict[Hashable, Number | dict[Hashable, Number]] | pd.Series,
    validate: bool = False,
    top: Number = 1,
    drop: bool = False,
) -> SeriesOrFrame:
    """
    Calculate the weighted score of selected columns in the DataFrame.

    The weighted score is the sum of the values in the DataFrame multiplied by
    the weights. A sugar syntax wraps::

        (df * weights).sum(axis=1) / sum(weights)

    Parameters
    ----------
    weights : list, dict or Series
        The weights of each column in the DataFrame.

        - list : The weights of each column in the DataFrame.
        - dict : Receive like ``{column: score}`` or ``{new_column: {column: score}}``.
        - Series : The weights must be a series with the same index as the DataFrame.

    validate : bool, default False
        If True, require the sum of ``weights`` values equal to 1.

    top : int, default 1
        If ``validate`` is True, require the sum of ``weights`` values equal to ``top``.

    drop : bool, default False
        If True, drop the used columns.

    Returns
    -------
    DataFrame or Series

    Raises
    ------
    TypeError
        - If one of the ``weights`` values is not number type.
        - If ``weights`` is not a list, a dict or a Series type.
        - If ``weights`` is a dict and the value is not a number or a dict type.

    ValueError
        - If ``weights`` is list type and its length is not the same as the number of
          DataFrame columns.
        - If ``weights`` is Series type and its labels are not in the DataFrame columns.
        - If ``weights`` is Series type and its labels are duplicated.
        - If ``validate=True`` and the sum of ``weights`` values is not equal to 1.

    See also
    --------
    pandas.DataFrame.mean : Calculate the mean of the DataFrame.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> df = pd.DataFrame({"a": [1, 1], "b": [2, 2], "c": [4, 4]})
    >>> df
       a  b  c
    0  1  2  4
    1  1  2  4

    Select all columns to calculate the score.

    >>> df.weighted_mean([0, 5, 5])
    0    3.0
    1    3.0
    dtype: float64

    Select some of columns and calculate the score.

    >>> df.weighted_mean({'b': 5, 'c': 5})
    0    3.0
    1    3.0
    dtype: float64

    Keep the original columns.

    >>> df.weighted_mean({"bc": {'b': 5, 'c': 5}})
       a  b   bc  c
    0  1  2  3.0  4
    1  1  2  3.0  4

    While ``weights`` is a dict and its values are also dict, it could use new generated
    columns to generate the score.

    >>> df.weighted_mean(
    ...     {
    ...         "ab": {"a": 1, "b": 1},
    ...         "bc": {"b": 1, "c": 1},
    ...         "ab-bc": {"ab": 1, "bc": 1},  # 'ab' and 'bc' are new generated columns
    ...     }
    ... )
       a   ab  ab-bc  b   bc  c
    0  1  1.5   2.25  2  3.0  4
    1  1  1.5   2.25  2  3.0  4
    """

    if isinstance(weights, (list, tuple)):
        result = score(df, weights, validate=validate, top=top)

    elif isinstance(weights, pd.Series):
        result = score(df, weights, validate=validate, top=top, name=weights.name)

    elif isinstance(weights, dict):
        if all(map(is_number, weights.values())):
            result = score(df, weights=weights, validate=validate, top=top)

        elif all(map(is_dict_like, weights.values())):
            result = pd.DataFrame()
            for name, weight in weights.items():
                if not all(map(is_number, weight.values())):
                    raise TypeError("The value of weights is not number type.")

                data = result.combine_first(df)
                another = score(data, weight, validate=validate, top=top, name=name)
                result = pd.concat((result, another), axis=1)

        else:
            raise TypeError("Received an invalid 'weights' type.")

    else:
        raise TypeError(
            "'weights' must be a list, a dict or a Series type, "
            f"but you passed a {type(weights).__name__!r}.",
        )

    if not drop:
        if isinstance(result, pd.Series) and result.name:
            result = result.to_frame()
        if isinstance(result, pd.DataFrame):
            result = result.combine_first(df)

    return result


def score(
    df: pd.DataFrame,
    /,
    weights: list[Number] | dict[Hashable, Number] | pd.Series,
    *,
    validate: bool,
    top: Number,
    name: Hashable = None,
) -> pd.Series:
    """Return calculated single score column."""

    if isinstance(weights, dict):
        df = df[weights.keys()]
        weights = weights.values()
    elif isinstance(weights, pd.Series):
        df = df[weights.index]
        weights = weights.values

    if validate and sum(weights) != top:
        raise ValueError(f"{sum(weights)=} is not equal to {top}.")

    return df.mul(weights).sum(axis=1).divide(sum(weights)).rename(name)



from __future__ import annotations

from typing import Hashable

import pandas as pd

from dtoolkit.accessor.register import register_index_method


@register_index_method
def to_set(index: pd.Index, /, level: int | Hashable = None) -> set:
    """
    Return a :keyword:`set` of the values.

    A sugary syntax wraps :keyword:`set`::

        set(index)

    Parameters
    ----------
    level : int or Hashable, optional
        Only return values from specified level (for :class:`~pandas.MultiIndex`).
        If int, gets the level by integer position, else by level name.

    Returns
    -------
    set

    See Also
    --------
    pandas.Index.unique
    dtoolkit.accessor.series.to_set

    Notes
    -----
    Different to :meth:`~pandas.Index.unique`, it returns :class:`~pandas.Index`.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> index = pd.Index([1, 2, 2])
    >>> index
    Int64Index([1, 2, 2], dtype='int64')
    >>> index.to_set()
    {1, 2}
    >>> index = pd.MultiIndex.from_arrays(
    ...     [
    ...         [1, 1, 2, 2],
    ...         ['red', 'blue', 'red', 'blue'],
    ...     ],
    ...     names=['number', 'color'],
    ... )
    >>> index
    MultiIndex([(1,  'red'),
                (1, 'blue'),
                (2,  'red'),
                (2, 'blue')],
               names=['number', 'color'])
    >>> index.to_set(0)
    {1, 2}
    >>> index.to_set('number')
    {1, 2}
    """

    return set(index.unique(level=level))



import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.register import register_geoseries_method


@register_geoseries_method
@doc(klass="GeoSeries")
def count_coordinates(s: gpd.GeoSeries, /) -> pd.Series:
    """
    Counts the number of coordinate pairs in each geometry of
    :class:`~geopandas.{klass}`.

    A sugary syntax wraps :meth:`pygeos.coordinates.count_coordinates`.

    Returns
    -------
    Series

    Raises
    ------
    ModuleNotFoundError
        If don't have module named 'pygeos'.

    See Also
    --------
    pygeos.coordinates.count_coordinates
        The core algorithm of this accessor.

    dtoolkit.geoaccessor.geoseries.count_coordinates
        Counts the number of coordinate pairs in each geometry of GeoSeries.

    dtoolkit.geoaccessor.geodataframe.count_coordinates
        Counts the number of coordinate pairs in each geometry of GeoDataFrame.

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import geopandas as gpd
    >>> df = gpd.GeoSeries.from_wkt(
    ...     [
    ...         "POINT (0 0)",
    ...         "LINESTRING (2 2, 4 4)",
    ...         None,
    ...     ],
    ... ).to_frame("geometry")
    >>> df
                                            geometry
    0                        POINT (0.00000 0.00000)
    1  LINESTRING (2.00000 2.00000, 4.00000 4.00000)
    2                                           None
    >>> df.count_coordinates()
    0    1
    1    2
    2    0
    Name: geometry, dtype: int64
    """

    try:
        from shapely import count_coordinates

        # NOTE: requires shapely>=2.0

        return s.apply(count_coordinates)

    except ImportError:
        # TODO: delete pygeos after shapely 2.x released
        from pygeos import count_coordinates, from_shapely

        return s.apply(lambda x: count_coordinates(from_shapely(x)))



from typing import Literal

import geopandas as gpd
import pandas as pd

from dtoolkit.accessor.series import set_unique_index  # noqa: F401
from dtoolkit.accessor.series import swap_index_values  # noqa: F401
from dtoolkit.geoaccessor.register import register_geoseries_method


BINARY_PREDICATE = Literal[
    "intersects",
    "crosses",
    "overlaps",
    "touches",
    "covered_by",
    "contains_properly",
    "contains",
    "within",
    "covers",
]


@register_geoseries_method
def duplicated_geometry_groups(
    s: gpd.GeoSeries,
    /,
    predicate: BINARY_PREDICATE = "intersects",
) -> pd.Series:
    """
    Labels of duplicate geometries.

    Parameters
    ----------
    predicate : {'intersects', 'crosses', 'overlaps', 'touches', 'covered_by', \
'contains_properly', 'contains', 'within', 'covers'}, default 'intersects'
        The binary predicate is used to validate whether the geometries are duplicates
        or not.

    Returns
    -------
    Series
        ``index`` is the index of inputting, ``values`` is the labels of groups.
        And labels are natural numbers.

    Warns
    -----
    UserWarning
        If the index of the inputting is not unique.

    See Also
    --------
    geopandas.sjoin
    dtoolkit.geoaccessor.geoseries.duplicated_geometry
    dtoolkit.geoaccessor.geoseries.duplicated_geometry_groups
    dtoolkit.geoaccessor.geodataframe.duplicated_geometry
    dtoolkit.geoaccessor.geodataframe.duplicated_geometry_groups

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import geopandas as gpd
    >>> from shapely.geometry import Polygon
    >>> df = gpd.GeoDataFrame(
    ...     geometry=[
    ...         Polygon([(0,0), (1,0), (1,1), (0,1)]),
    ...         Polygon([(1,1), (2,1), (2,2), (1,2)]),
    ...         Polygon([(2,2), (3,2), (3,3), (2,3)]),
    ...         Polygon([(2, 0), (3, 0), (3, 1)]),
    ...     ],
    ... )
    >>> df
                                                geometry
    0  POLYGON ((0.00000 0.00000, 1.00000 0.00000, 1....
    1  POLYGON ((1.00000 1.00000, 2.00000 1.00000, 2....
    2  POLYGON ((2.00000 2.00000, 3.00000 2.00000, 3....
    3  POLYGON ((2.00000 0.00000, 3.00000 0.00000, 3....

    - 0 and 1 are intersecting.
    - 1 and 2 are intersecting.
    - 3 is alone.

    So there are two groups: ``(0, 1, 2)`` and ``(3,)``.

    >>> df.duplicated_geometry_groups()
    0    0
    1    0
    2    0
    3    1
    dtype: int64
    """

    # NOTE: `predicate` requires geopandas >= 0.10.0
    return (
        s.to_frame("geometry")
        .set_unique_index(drop=True)
        .pipe(self_sjoin, predicate=predicate)
        .drop_geometry()
        .to_series()
        .pipe(group_shared_xy, size=s.size)
        .set_axis(s.index)
    )


def self_sjoin(df: gpd.GeoDataFrame, /, **kwargs) -> gpd.GeoDataFrame:
    """
    Perform self-sjoin on a GeoDataFrame.

    Parameters
    ----------
    **kwargs
        See the documentation for :meth:`~geopandas.sjoin` for complete
        details on the keyword arguments except ``left_df`` and ``right_df``.

    Returns
    -------
    GeoDataFrame

    See Also
    --------
    geopandas.sjoin

    Notes
    -----
    ``left_df`` and ``right_df`` parameter are ``df`` and can't be overridden.
    """

    return gpd.sjoin(df, df, **kwargs)


def group_shared_xy(s: pd.Series, /, size: int) -> pd.Series:
    """
    Group the shared 'x' and 'y' coordinates.

    A simple example::

        # group x-y relationship

        # diagonal matrix in logical
        #    0  1  2  3
        # 0  √
        # 1     √  √
        # 2     √  √  √
        # 3        √  √

        # which data in real
        [
            (1, 2),  # or `(2, 1)`
            (2, 3),  # or `(3, 2)`
        ]

        # into

        [
            {0},
            {1, 2, 3},
        ]

    Parameters
    ----------
    s : Series
        Its index is the 'x' and values is the 'y'. Require all 'x' value is not equal
        to 'y' value 'x-y' (or 'y-x') parts.

    size : int
        The size of the matrix.

    Returns
    -------
    Series
        index is the original ``index``, ``values`` is the label of group.

    Notes
    -----
    - Only support 2d coordinates, it means data unit is ``(x, y)``.
    - Only support 'diagonal maxtrix', it means 'x' value equal to 'y' value.
    - Only support natural numbers of labels.

    Examples
    --------
    >>> import pandas as pd
    >>> s = pd.Series([2, 3], name='y', index=pd.Index([1, 2], name='x'))
    >>> s
    x
    1    2
    2    3
    Name: y, dtype: int64
    >>> group_shared_xy(s, 4)
    0    1
    1    0
    2    0
    3    0
    dtype: int64
    """

    groups = []
    for x, y in s.items():
        for group in groups:
            if x in group or y in group:
                group.update({x, y})
                break
        else:
            groups.append({x, y})

    # add missing group via natural number
    for i in range(size):
        for group in groups:
            if i in group:
                break
        else:
            groups.append({i})

    return pd.Series(groups).explode().swap_index_values().sort_index()



import geopandas as gpd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries import cncrs_offset as s_cncrs_offset
from dtoolkit.geoaccessor.geoseries.cncrs_offset import CHINA_CRS
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_cncrs_offset, klass="GeoDataFrame")
def cncrs_offset(
    df: gpd.GeoDataFrame,
    /,
    from_crs: CHINA_CRS,
    to_crs: CHINA_CRS,
) -> gpd.GeoDataFrame:

    return df.assign(
        **{
            df.geometry.name: s_cncrs_offset(
                df.geometry,
                from_crs=from_crs,
                to_crs=to_crs,
            ),
        }
    )



from typing import Literal
from typing import Union

import numpy as np
import pandas as pd


SeriesOrFrame = Union[pd.Series, pd.DataFrame]

TwoDimArray = Union[np.ndarray, pd.DataFrame]
OneDimArray = Union[np.ndarray, pd.Series]

Number = Union[int, float]

Axis = Literal[0, 1, "index", "columns"]  # only for dataframe axis



from typing import Hashable

import geopandas as gpd
import pandas as pd

from dtoolkit.accessor.dataframe import drop_or_not
from dtoolkit.accessor.register import register_dataframe_method


@register_dataframe_method
def geocode(
    df: pd.DataFrame,
    /,
    address: Hashable,
    drop: bool = False,
    **kwargs,
) -> gpd.GeoDataFrame:
    """
    Geocode a string type column from a DataFrame and get a GeoDataFrame of the
    resulting points.

    Parameters
    ----------
    address : Hashable
        The name of the column to geocode.

    drop : bool, default False
        Don't contain the original data anymore.

    **kwargs
        See the documentation for :func:`~geopandas.tools.geocode` for complete details
        on the keyword arguments.

    Returns
    -------
    GeoDataFrame

    Raises
    ------
    ModuleNotFoundError
        If don't have module named 'geopy'.

    See Also
    --------
    geopandas.tools.geocode
    dtoolkit.geoaccessor.series.geocode

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> df = pd.DataFrame(
    ...     {
    ...         "name": [
    ...             "boston, ma",
    ...             "1600 pennsylvania ave. washington, dc",
    ...         ],
    ...     }
    ... )
    >>> df
                                        name
    0                             boston, ma
    1  1600 pennsylvania ave. washington, dc
    >>> df.geocode("name", drop=True)
                         geometry                                            address
    0  POINT (-71.06051 42.35543)               Boston, Massachusetts, United States
    1  POINT (-77.03655 38.89770)  White House, 1600, Pennsylvania Avenue Northwe...
    """

    return pd.concat(
        (
            gpd.tools.geocode(df[address], **kwargs),
            drop_or_not(df, drop=drop, columns=address),
        ),
        axis=1,
    )



from __future__ import annotations

import pandas as pd

from dtoolkit._typing import Number
from dtoolkit._typing import OneDimArray
from dtoolkit.accessor.register import register_series_method
from dtoolkit.util._decorator import warning


@register_series_method
@warning(
    "The keyword argument 'columns' is deprecated, "
    "please use 'absolute_error' and 'relative_error' instead. "
    "(Warning added DToolKit 0.0.19)",
    category=DeprecationWarning,
    stacklevel=3,
)
def error_report(
    s: pd.Series,
    predicted: OneDimArray | list[Number],
    /,
    absolute_error: str = "absolute_error",
    relative_error: str = "relative_error",
) -> pd.DataFrame:
    """
    Calculate `absolute_error` and `relative_error` of two columns.

    .. math::

        absolute\\_error = \\lvert predicted - s \\rvert

        relative\\_error = \\frac{absolute\\_error}{s}

    Parameters
    ----------
    predicted : list of int or float, ndarrray, Series
        A array is compared to ``s``.

    columns : list of Hashable, optional
        The columns of returning DataFrame, each represents `true value`,
        `predicted value`, `absolute error`, and `relative error`.

        .. deprecated:: 0.0.19
            Please use 'absolute_error' and 'relative_error' instead.

    absolute_error : str, default 'absolute_error'
        The name of the column of absolute error.

    relative_error : str, default 'relative_error'
        The name of the column of relative error.

    Returns
    -------
    DataFrame
        Return four columns DataFrame and each represents 'true value',
        'predicted value', 'absolute error', and 'relative error'.

    Raises
    ------
    IndexError
        - If ``len(s)`` != ``len(predicted)``.
        - If ``predicted`` is Series and its index not equal to ``s``'s index.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series([1, 2, 3])
    >>> s.error_report([3, 2, 1])
       true  predicted  absolute_error  relative_error
    0     1          3               2        2.000000
    1     2          2               0        0.000000
    2     3          1               2        0.666667

    If the name of ``s`` or ``predicted`` is not None, the columns of
    ``error_report`` would use the name of ``s`` and ``predicted``.

    >>> s = pd.Series([1, 2, 3], name="y")
    >>> predicted = pd.Series([3, 2, 1], name="y predicted")
    >>> s.error_report(predicted)
       y  y predicted  absolute_error  relative_error
    0  1            3               2        2.000000
    1  2            2               0        0.000000
    2  3            1               2        0.666667

    Set ``absolute_error`` and ``relative_error``.

    >>> s.rename('a').error_report(
    ...     predicted.rename('b'),
    ...     absolute_error="c",
    ...     relative_error="d",
    ... )
       a  b  c         d
    0  1  3  2  2.000000
    1  2  2  0  0.000000
    2  3  1  2  0.666667
    """

    if s.size != len(predicted):
        raise IndexError(
            "Length of 'predicted' doesn't match length of 'reference'.",
        )

    if isinstance(predicted, pd.Series):
        if not s.index.equals(predicted.index):
            raise IndexError(
                "Index values of 'predicted' sequence doesn't "
                "match index values of 'reference'.",
            )
    else:
        predicted = pd.Series(predicted, index=s.index)

    absolute = (predicted - s).abs()
    relative = absolute / s
    return pd.concat(
        (
            s.rename(s.name or "true"),
            predicted.rename(predicted.name or "predicted"),
            absolute.rename(absolute_error),
            relative.rename(relative_error),
        ),
        axis=1,
    )



from __future__ import annotations

from typing import Literal

import geopandas as gpd
import pandas as pd

from dtoolkit.geoaccessor.geoseries.duplicated_geometry_groups import BINARY_PREDICATE
from dtoolkit.geoaccessor.geoseries.duplicated_geometry_groups import (
    duplicated_geometry_groups,
)
from dtoolkit.geoaccessor.register import register_geoseries_method
from dtoolkit.util._decorator import warning


@register_geoseries_method
@warning(
    "`Geo(Series|DataFrame).duplicated_geometry`'s `predicate` default value is "
    "changed from `'intersects'` to `None`. (Warning added DToolKit 0.0.19)",
)
def duplicated_geometry(
    s: gpd.GeoDataFrame,
    /,
    predicate: BINARY_PREDICATE | None = None,
    keep: Literal["first", "last", False] = "first",
) -> pd.Series:
    """
    Return boolean Series denoting duplicate geometries.

    Parameters
    ----------
    predicate : {'intersects', 'crosses', 'overlaps', 'touches', 'covered_by', \
'contains_properly', 'contains', 'within', 'covers', None}, default None
        The binary predicate is used to validate whether the geometries are duplicates
        or not. If None, the geometries will directly compares via value relation
        instead of the spatial relation.

        .. versionchanged:: 0.0.19
            The 'predicate' default value is changed from 'intersects' to None.

    keep : {'first', 'last', False}, default 'first'
        - ``first`` : Mark duplicates as ``True`` except for the first occurrence.
        - ``last`` : Mark duplicates as ``True`` except for the last occurrence.
        - False : Mark all duplicates as ``True``.

    Returns
    -------
    Series

    See Also
    --------
    pandas.Series.duplicated
    geopandas.sjoin
    dtoolkit.geoaccessor.geoseries.duplicated_geometry
    dtoolkit.geoaccessor.geodataframe.duplicated_geometry

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import geopandas as gpd

    Drop duplicated geometries by value equal.

    >>> from shapely.geometry import Point
    >>> df = gpd.GeoDataFrame(geometry=[Point(0, 0), Point(0, 0), Point(1, 1)])
    >>> df
                      geometry
    0  POINT (0.00000 0.00000)
    1  POINT (0.00000 0.00000)
    2  POINT (1.00000 1.00000)
    >>> df.duplicated_geometry()
    0    False
    1     True
    2    False
    Name: geometry, dtype: bool

    Drop duplicated geometries by sptial relation.

    >>> from shapely.geometry import Polygon
    >>> df = gpd.GeoDataFrame(
    ...     geometry=[
    ...         Polygon([(0,0), (1,0), (1,1), (0,1)]),
    ...         Polygon([(1,1), (2,1), (2,2), (1,2)]),
    ...         Polygon([(2,2), (3,2), (3,3), (2,3)]),
    ...         Polygon([(2, 0), (3, 0), (3, 1)]),
    ...     ],
    ... )
    >>> df
                                                geometry
    0  POLYGON ((0.00000 0.00000, 1.00000 0.00000, 1....
    1  POLYGON ((1.00000 1.00000, 2.00000 1.00000, 2....
    2  POLYGON ((2.00000 2.00000, 3.00000 2.00000, 3....
    3  POLYGON ((2.00000 0.00000, 3.00000 0.00000, 3....
    >>> df.duplicated_geometry('intersects')
    0    False
    1     True
    2     True
    3    False
    dtype: bool
    """

    if predicate is not None:
        s = duplicated_geometry_groups(s, predicate=predicate)
    return s.duplicated(keep=keep)



from dtoolkit.geoaccessor.geodataframe.cncrs_offset import cncrs_offset  # noqa: F401
from dtoolkit.geoaccessor.geodataframe.count_coordinates import (  # noqa: F401
    count_coordinates,
)
from dtoolkit.geoaccessor.geodataframe.drop_duplicates_geometry import (  # noqa: F401
    drop_duplicates_geometry,
)
from dtoolkit.geoaccessor.geodataframe.drop_geometry import drop_geometry  # noqa: F401
from dtoolkit.geoaccessor.geodataframe.duplicated_geometry import (  # noqa: F401
    duplicated_geometry,
)
from dtoolkit.geoaccessor.geodataframe.duplicated_geometry_groups import (  # noqa: F401
    duplicated_geometry_groups,
)
from dtoolkit.geoaccessor.geodataframe.filter_geometry import (  # noqa: F401
    filter_geometry,
)
from dtoolkit.geoaccessor.geodataframe.geoarea import geoarea  # noqa: F401
from dtoolkit.geoaccessor.geodataframe.geobuffer import geobuffer  # noqa: F401
from dtoolkit.geoaccessor.geodataframe.geodistance import geodistance  # noqa: F401
from dtoolkit.geoaccessor.geodataframe.geodistance_matrix import (  # noqa: F401
    geodistance_matrix,
)
from dtoolkit.geoaccessor.geodataframe.get_coordinates import (  # noqa: F401
    get_coordinates,
)
from dtoolkit.geoaccessor.geodataframe.has_hole import has_hole  # noqa: F401
from dtoolkit.geoaccessor.geodataframe.hole_counts import hole_counts  # noqa: F401
from dtoolkit.geoaccessor.geodataframe.reverse_geocode import (  # noqa: F401
    reverse_geocode,
)
from dtoolkit.geoaccessor.geodataframe.toposimplify import toposimplify  # noqa: F401



import pandas as pd
from pandas.core.computation.eval import eval as _eval

from dtoolkit.accessor.register import register_series_method


@register_series_method
def eval(s: pd.Series, /, expr: str, **kwargs):
    """
    Evaluate a string describing operations on Series.

    This allows ``eval`` to run arbitrary code, which can make you vulnerable to code
    injection if you pass user input to this function.

    Parameters
    ----------
    expr : str
        The expression string to evaluate.

    **kwargs
        See the documentation for :meth:`pandas.eval` for complete details on
        the keyword arguments.

    Returns
    -------
    ndarray, scalar, pandas object

    See Also
    --------
    pandas.eval
        Evaluate a Python expression as a string using various backends.

    pandas.DataFrame.eval
        Evaluate a string describing operations on DataFrame columns.

    dtoolkit.accessor.series.query
        Evaluates a boolean expression to query Series.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd

    The ``name`` should be set.

    >>> s = pd.Series([1, 2, 3], index=["a", "b", "c"], name="col")
    >>> s
    a    1
    b    2
    c    3
    Name: col, dtype: int64
    >>> s.eval("col - 1")
    a    0
    b    1
    c    2
    Name: col, dtype: int64
    >>> s.eval("d = 4")
    a    1
    b    2
    c    3
    d    4
    Name: col, dtype: int64
    >>> s.eval("col == 1")
    a     True
    b    False
    c    False
    Name: col, dtype: bool
    >>> s.eval("index == 'c'")
    a    False
    b    False
    c     True
    dtype: bool
    """

    index_resolvers = s._get_index_resolvers()
    column_resolvers = s._get_cleaned_column_resolvers()
    resolvers = column_resolvers, index_resolvers

    kwargs["level"] = kwargs.pop("level", 0) + 1
    kwargs["target"] = kwargs.get("target", s)
    kwargs["resolvers"] = tuple(kwargs.get("resolvers", ())) + resolvers

    return _eval(expr, **kwargs)



from __future__ import annotations

from typing import get_args
from typing import Literal
from typing import TYPE_CHECKING

import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.register import register_geoseries_method


if TYPE_CHECKING:
    from shapely.geometry.base import BaseGeometry


BINARY_PREDICATE = Literal[
    "intersects",
    "disjoint",
    "crosses",
    "overlaps",
    "touches",
    "covered_by",
    "contains",
    "within",
    "covers",
]


@register_geoseries_method
@doc(klass="GeoSeries")
def filter_geometry(
    s: gpd.GeoSeries,
    /,
    other: BaseGeometry | gpd.GeoSeries | gpd.GeoDataFrame,
    predicate: BINARY_PREDICATE,
    complement: bool = False,
    **kwargs,
) -> gpd.GeoSeries:
    """
    Filter {klass} via the spatial relationship between {klass} and ``geometry``.

    A sugar syntax wraps::

        s[s.{{predicate}}(other, **kwargs)]

    Parameters
    ----------
    other : Geometry, GeoSeries, or GeoDataFrame

    predicate : {{"intersects", "disjoint", "crosses", "overlaps", "touches", \
"covered_by", "contains", "within", "covers"}}
        Binary predicate.

    complement : bool, default False
        If True, do operation reversely.

    **kwargs
        See the documentation for ``{klass}.{{predicate}}`` for complete details on the
        keyword arguments.

    Returns
    -------
    {klass}

    See Also
    --------
    geopandas.GeoSeries.intersects
    geopandas.GeoSeries.covered_by
    geopandas.GeoSeries.contains
    dtoolkit.geoaccessor.geoseries.filter_geometry
    dtoolkit.geoaccessor.geodataframe.filter_geometry

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import geopandas as gpd
    >>> from shapely.geometry import Polygon, LineString, Point, box
    >>> df = gpd.GeoDataFrame(
    ...     geometry=[
    ...         Polygon([(0, 0), (1, 1), (0, 1)]),
    ...         LineString([(0, 0), (0, 2)]),
    ...         LineString([(0, 0), (0, 1)]),
    ...         Point(0, 1),
    ...         Point(-1, -1),
    ...     ],
    ... )
    >>> df
                                                geometry
    0  POLYGON ((0.00000 0.00000, 1.00000 1.00000, 0....
    1      LINESTRING (0.00000 0.00000, 0.00000 2.00000)
    2      LINESTRING (0.00000 0.00000, 0.00000 1.00000)
    3                            POINT (0.00000 1.00000)
    4                          POINT (-1.00000 -1.00000)

    Filter the geometries out of the bounding ``box(0, 0, 2, 2)``.

    >>> df.filter_geometry(box(0, 0, 2, 2), "covered_by", complement=True)
                          geometry
    4    POINT (-1.00000 -1.00000)

    This method is actually faster than the following one::

        def is_in_geometry(s: gpd.GeoSeries, other: BaseGeometry) -> pd.Series:
            s_bounds, g_bounds = s.bounds, other.bounds

            return (
                (s_bounds.minx >= g_bounds[0])
                & (s_bounds.miny >= g_bounds[1])
                & (s_bounds.maxx <= g_bounds[2])
                & (s_bounds.maxy <= g_bounds[3])
            )
    """

    return s[
        _filter_geometry(
            s,
            other=other,
            predicate=predicate,
            complement=complement,
            **kwargs,
        )
    ]


def _filter_geometry(
    s: gpd.GeoSeries,
    other: BaseGeometry | gpd.GeoSeries | gpd.GeoDataFrame,
    predicate: BINARY_PREDICATE,
    complement: bool,
    **kwargs,
) -> pd.Series:
    if predicate not in get_args(BINARY_PREDICATE):
        raise ValueError(
            f"Got {predicate=!r}, expected one of {get_args(BINARY_PREDICATE)!r}.",
        )

    mask = getattr(s, predicate)(other, **kwargs)
    return ~mask if complement else mask



from __future__ import annotations

from functools import wraps
from warnings import warn

from dtoolkit.util._exception import find_stack_level


def deprecated_kwargs(
    *arguments: list[str],
    message: str = "The keyword argument '{argument}' of '{func_name}' is deprecated.",
):
    """
    Used as a decorator when deprecating function keyword argument names.

    Parameters
    -----------
    arguments : list of str
        The list of deprecated keyword arguments.

    message : str
        The message of :exc:`DeprecationWarning`. It should be a string or a
        string template. If a string template defaults input ``func_name`` and
        ``argument``.

    Raises
    ------
    DeprecationWarning
        If one of `arguments` is input into the function.

    See Also
    --------
    dtoolkit.util._decorator.warning
    dtoolkit.util._decorator.deprecated_alias

    Examples
    --------
    >>> from dtoolkit.util._decorator import deprecated_kwargs
    >>> @deprecated_kwargs('x', 'y')
    ... def plus(a, b, x=0, y=0):
    ...     return a + b
    >>> plus(1, 2, x=1, y=2)
    3
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for argument in arguments:
                if argument in kwargs:
                    warn(
                        message.format(func_name=func.__name__, argument=argument),
                        DeprecationWarning,
                        find_stack_level(),
                    )

            return func(*args, **kwargs)

        return wrapper

    return decorator



import pandas as pd

from dtoolkit.transformer import DataFrameTF


class AssignTF(DataFrameTF):
    """
    A transformer could assign new columns to a :obj:`~pandas.DataFrame`.

    See Also
    --------
    pandas.DataFrame.assign : This transformer's prototype method.

    Examples
    --------
    >>> import pandas as pd
    >>> from dtoolkit.transformer import AssignTF
    >>> df = pd.DataFrame({'temp_c': [17.0, 25.0]}, index=['Portland', 'Berkeley'])
    >>> df
                temp_c
    Portland    17.0
    Berkeley    25.0

    Where the value is a callable, evaluated on ``df``:

    >>> pipeline = AssignTF(temp_f=lambda x: x.temp_c * 9 / 5 + 32)
    >>> pipeline.transform(df)
                temp_c  temp_f
    Portland    17.0    62.6
    Berkeley    25.0    77.0
    """

    transform_method = staticmethod(pd.DataFrame.assign)



import pandas as pd

from dtoolkit.accessor.register import register_dataframe_method


@register_dataframe_method
def groupby_index(df: pd.DataFrame, /, **kwargs) -> pd.core.groupby.DataFrameGroupBy:
    """
    Group DataFrame by its index.

    Parameters
    ----------
    **kwargs
        See the documentation for :meth:`~pandas.DataFrame.groupby` for complete
        details on the keyword arguments except ``by``.

    Returns
    -------
    DataFrameGroupBy

    See Also
    --------
    pandas.DataFrame.groupby
    dtoolkit.accessor.series.groupby_index

    Notes
    -----
    ``by`` parameter is set to ``df.index`` and can't be overridden.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> df = pd.DataFrame({'a': [1, 1, 1], 'b': [1, 2, 3]}, index=['a', 'b', 'a'])
    >>> df
       a  b
    a  1  1
    b  1  2
    a  1  3
    >>> df.groupby_index().sum()
       a  b
    a  2  4
    b  1  2
    """

    return df.groupby(df.index, **kwargs)



import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geodataframe import drop_geometry
from dtoolkit.geoaccessor.geoseries import reverse_geocode as s_reverse_geocode
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_reverse_geocode, klass="GeoDataFrame")
def reverse_geocode(df: gpd.GeoDataFrame, /, **kwargs) -> gpd.GeoDataFrame:

    return pd.concat(
        (
            s_reverse_geocode(df.geometry, **kwargs),
            drop_geometry(df),
        ),
        axis=1,
    )



import pandas as pd
from pandas.util._decorators import doc

from dtoolkit._typing import Axis
from dtoolkit.accessor.register import register_dataframe_method
from dtoolkit.accessor.series.change_axis_type import (
    change_axis_type as s_change_axis_type,
)


@register_dataframe_method
@doc(s_change_axis_type, alias="df", klass="DataFrame")
def change_axis_type(df: pd.DataFrame, dtype: type, axis: Axis = 0) -> pd.DataFrame:
    axis = df._get_axis_number(axis)

    df = df.copy()  # Avoid mutating the original DataFrame
    if axis == 0:
        df.index = df.index.astype(dtype)
    else:
        df.columns = df.columns.astype(dtype)

    return df



from __future__ import annotations

from typing import Literal

import geopandas as gpd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries import toposimplify as s_toposimplify
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_toposimplify, klass="GeoDataFrame")
def toposimplify(
    df: gpd.GeoDataFrame,
    /,
    tolerance: float,
    simplify_algorithm: Literal["dp", "vw"] = "dp",
    simplify_with: Literal["shapely", "simplification"] = "shapely",
    prevent_oversimplify: bool = True,
) -> gpd.GeoDataFrame:

    return df.assign(
        **{
            df.geometry.name: s_toposimplify(
                df.geometry,
                tolerance=tolerance,
                simplify_algorithm=simplify_algorithm,
                simplify_with=simplify_with,
                prevent_oversimplify=prevent_oversimplify,
            ),
        }
    )



import pandas as pd

from dtoolkit.accessor.register import register_series_method


@register_series_method
def query(s: pd.Series, /, expr: str, **kwargs) -> pd.Series:
    """
    Query the columns of a Series with a boolean expression.

    Parameters
    ----------
    expr : str
        The query string to evaluate.

    **kwargs
        See the documentation for :meth:`pandas.eval` for complete details on
        the keyword arguments.

    Returns
    -------
    Series

    See Also
    --------
    pandas.eval
        Evaluate a Python expression as a string using various backends.

    pandas.DataFrame.query
        Evaluates a boolean expression to query DataFrame.

    dtoolkit.accessor.series.eval
        Evaluate a string describing operations on Series columns.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd

    The ``name`` should be set.

    >>> s = pd.Series([1, 2, 3], index=["a", "b", "c"], name="col")
    >>> s
    a    1
    b    2
    c    3
    Name: col, dtype: int64
    >>> s.query("col == 1")
    a    1
    Name: col, dtype: int64
    >>> s.query("index == 'c'")
    c    3
    Name: col, dtype: int64
    """

    if not isinstance(expr, str):
        raise ValueError(
            f"'expr' must be a string to be evaluated, {type(expr)} given",
        )

    kwargs["level"] = kwargs.pop("level", 0) + 1
    kwargs["target"] = None
    mask = s.eval(expr, **kwargs)

    return s[mask]



import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries import get_coordinates as s_get_coordinates
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_get_coordinates, klass="GeoDataFrame")
def get_coordinates(df: gpd.GeoDataFrame, /, **kwargs) -> pd.Series:
    return s_get_coordinates(df.geometry, **kwargs)



from warnings import catch_warnings
from warnings import simplefilter
from warnings import warn

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

from dtoolkit._typing import Axis
from dtoolkit.accessor.register import register_dataframe_method
from dtoolkit.util._exception import find_stack_level


@register_dataframe_method
def equal(
    df: pd.DataFrame,
    /,
    other,
    align: bool = True,
    axis: Axis = 0,
    **kwargs,
) -> pd.DataFrame:
    """
    Return a boolean DataFrame containing the result of comparing with ``other``.

    A sugar syntax for ``np.equal(df, other, **kwargs)``.

    Parameters
    ----------
    other : scalar, 1d array-like or 2d array-like
        The value(s) to compare with the DataFrame.

        - scalar : compare each element with the scalar value.
        - 1d / 2d array-like : compare each element with the corresponding value.

    align : bool, default True
        If True, align ``df`` and ``other`` before comparing. Only works while
        ``other`` is a Series / DataFrame.

    axis : {0 or 'index', 1 or 'columns'}, default 0
        If 0, compare along ``df``'s rows else columns. Works only when ``other`` is
        1d array-like.

    **kwargs
        Additional keyword arguments are passed to :meth:`numpy.equal`.

    Returns
    -------
    DataFrame

    Warns
    -----
    UserWarning
        If ``align`` is True and ``other`` is a Series / DataFrame, but its
        indexes (columns) are not equal to ``df``.

    Raises
    ------
    ValueError
        If ``other`` is id array-like and its size is not equal to ``df``'s size.

    See Also
    --------
    pandas.DataFrame.eq : Compare two Series.
    numpy.equal : Return (x1 == x2) element-wise.
    dtoolkit.accessor.series.equal : Series version of ``.equal``.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> df = pd.DataFrame({"a": [1, 2, 3], "b": [3, 2, 1]})
    >>> df
       a  b
    0  1  3
    1  2  2
    2  3  1

    Like ``DataFrame.eq`` method to compare scalar or array-like via ``==``,
    ``DataFrame.equal`` could also do that.

    Compare with a scalar.

    >>> df == 1
           a      b
    0   True  False
    1  False  False
    2  False   True
    >>> df.equal(1)
           a      b
    0   True  False
    1  False  False
    2  False   True

    Compare with a 1d array-like.

    But ``==`` only can compare along the row. ``DataFrame.equal`` can compare along
    the row or column.

    >>> df == [1, 2]
           a      b
    0   True  False
    1  False   True
    2  False  False
    >>> df.equal([1, 2], axis=0)  # default compare along the row
           a      b
    0   True  False
    1  False   True
    2  False  False
    >>> df.equal([1, 2, 3], axis=1)  # compare along the column
          a      b
    0  True  False
    1  True   True
    2  True  False

    Compare with a 2d array-like.

    >>> import numpy as np
    >>> df == np.array([[1, 1], [2, 2], [3, 3]])
          a      b
    0  True  False
    1  True   True
    2  True  False
    >>> df.equal(np.array([[1, 1], [2, 2], [3, 3]]))  # pure list is also okay
          a      b
    0  True  False
    1  True   True
    2  True  False
    """

    axis = df._get_axis_number(axis)
    if align and (
        # For Series, `df.index` or `df.columns` must be equal to `other.index`.
        isinstance(other, pd.Series)
        and (
            axis == 0
            and not df.columns.equals(other.index)
            or axis == 1
            and not df.index.equals(other.index)
        )
        # For DataFrame, both `df.index` and `df.columns` must be equal to `other.index`
        or isinstance(other, pd.DataFrame)
        and not (df.index.equals(other.index) and df.columns.equals(other.columns))
    ):
        warn("the indices are different.", stacklevel=find_stack_level())
        a = 1 - axis if isinstance(other, pd.Series) else None
        df, other = df.align(other, axis=a)

    if is_list_like(other):
        other = np.asarray(other)
        if other.ndim == 1 and other.size != df.shape[1 - axis]:
            raise ValueError(
                f"{other.size=} does not equal to {df.shape[1 - axis]=}.",
            )

        if other.ndim == 1 and axis == 1:
            other = other.reshape((-1, 1))

    with catch_warnings():
        # TODO: remove `catch_warnings` in the future.
        # The latest pandas version (1.4.x) doesn't specify a version when
        # it will change the behavior on non-aligned DataFrame.

        # Original message:
        # FutureWarning: Calling a ufunc on non-aligned DataFrames(or DataFrame/Series
        # combination). Currently, the indices are ignored and the result takes the
        # index/columns of the first DataFrame. In the future , the DataFrames/Series
        # will be aligned before applying the ufunc. Convert one of the arguments to a
        # NumPy array (eg 'ufunc(df1, np.asarray(df2)') to keep the current behaviour,
        # or align manually (eg 'df1, df2 = df1.align(df2)') before passing to the ufunc
        # to obtain the future behaviour and silence this warning.
        simplefilter("ignore", FutureWarning)

        # NOTE: Requires pandas >= 1.2.0 to better support numpy ufunc.
        return np.equal(df, other, **kwargs)



import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries import has_hole as s_has_hole
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_has_hole)
def has_hole(df: gpd.GeoDataFrame, /) -> pd.Series:
    return s_has_hole(df.geometry)



import geopandas as gpd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.register import register_geoseries_method


@register_geoseries_method
@doc(klass="GeoSeries")
def reverse_geocode(s: gpd.GeoSeries, /, **kwargs) -> gpd.GeoDataFrame:
    """
    Reverse geocode :obj:`~shapely.geometry.Point` type :class:`~geopandas.{klass}`
    and get the corresponding addresses.

    Parameters
    ----------
    **kwargs
        See the documentation for :func:`~geopandas.tools.reverse_geocode` for complete
        details on the keyword arguments.

    Returns
    -------
    GeoDataFrame

    Raises
    ------
    ModuleNotFoundError
        If don't have module named 'geopy'.

    See Also
    --------
    geopandas.tools.reverse_geocode
    dtoolkit.geoaccessor.series.geocode
    dtoolkit.geoaccessor.dataframe.geocode
    dtoolkit.geoaccessor.geoseries.reverse_geocode
    dtoolkit.geoaccessor.geodataframe.reverse_geocode

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> df = (
    ...     pd.Series(
    ...         [
    ...             "POINT (-71.0594869 42.3584697)",
    ...             "POINT (-77.0365305 38.8977332)",
    ...         ],
    ...     )
    ...     .from_wkt(drop=True, crs=4326)
    ...     .to_frame("geometry")
    ... )
    >>> df
                         geometry
    0  POINT (-71.05949 42.35847)
    1  POINT (-77.03653 38.89773)
    >>> df.reverse_geocode()
                         geometry                                            address
    0  POINT (-71.05977 42.35860)  18-32, Tremont Street, 02108, Tremont Street, ...
    1  POINT (-77.03655 38.89772)  Pennsylvania Avenue Northwest, 20006, Pennsylv...
    """

    return gpd.tools.reverse_geocode(s.geometry, **kwargs)



from __future__ import annotations

from typing import Literal

import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries import duplicated_geometry as s_duplicated_geometry
from dtoolkit.geoaccessor.geoseries.duplicated_geometry_groups import BINARY_PREDICATE
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_duplicated_geometry)
def duplicated_geometry(
    df: gpd.GeoDataFrame,
    /,
    predicate: BINARY_PREDICATE | None = None,
    keep: Literal["first", "last", False] = "first",
) -> pd.Series:
    return s_duplicated_geometry(df.geometry, predicate=predicate, keep=keep)



from __future__ import annotations

from typing import Hashable
from typing import Iterable
from typing import Literal

import pandas as pd

from dtoolkit._typing import SeriesOrFrame
from dtoolkit.accessor.dataframe import boolean  # noqa: F401
from dtoolkit.accessor.register import register_dataframe_method


@register_dataframe_method
def filter_in(
    df: pd.DataFrame,
    condition: Iterable | SeriesOrFrame | dict[Hashable, list[Hashable]],
    /,
    how: Literal["any", "all"] = "all",
    complement: bool = False,
) -> pd.DataFrame:
    """
    Filter :obj:`~pandas.DataFrame` contents.

    Similar to :meth:`~pandas.DataFrame.isin`, but the return is not bool.

    Parameters
    ----------
    condition : Iterable, Series, DataFrame or dict
        The filtered result is based on this specific condition.

        * If ``condition`` is a :obj:`dict`, the keys must be the column
          names, which must be matched. And ``how`` only works on these gave keys.

        * If ``condition`` is a :obj:`~pandas.Series`, that's the index.

        * If ``condition`` is a :obj:`~pandas.DataFrame`, then both the index
          and column labels must be matched.

    how : {'any', 'all'}, default 'all'
        Determine whether the row is filtered from :obj:`~pandas.DataFrame`,
        when there have at least one value or all value.

        * 'any' : If any values are present, filter that rows.
        * 'all' : If all values are present, filter that rows.

    complement : bool, default is False
        If True, do operation reversely.

    Returns
    -------
    DataFrame

    See Also
    --------
    pandas.DataFrame.isin
        Whether each element in the DataFrame is contained in values.

    pandas.DataFrame.filter
        Subset the dataframe rows or columns according to the specified index
        labels.

    dtoolkit.accessor.series.filter_in
        Filter Series contents.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> df = pd.DataFrame(
    ...     {
    ...         'legs': [2, 4, 2],
    ...         'wings': [2, 0, 0],
    ...     },
    ...     index=['falcon', 'dog', 'cat'],
    ... )
    >>> df
            legs  wings
    falcon     2      2
    dog        4      0
    cat        2      0

    When ``condition`` is a list check whether every value in the DataFrame is
    present in the list (which animals have 0 or 2 legs or wings).

    Filter rows.

    >>> df.filter_in([0, 2])
            legs  wings
    falcon     2      2
    cat        2      0

    Filter any row doesn't contain 0 or 2.

    >>> df.filter_in([0, 2], how="any", complement=True)
            legs  wings
    dog        4      0

    When ``condition`` is a :obj:`dict`, we can pass values to check for each
    column separately.

    >>> df.filter_in({'legs': [2], 'wings': [2]})
            legs  wings
    falcon     2      2

    When ``values`` is a Series or DataFrame the index and column must be matched.
    Note that 'spider' doesn't match based on the number of legs in ``other``.

    >>> other = pd.DataFrame(
    ...     {
    ...         'legs': [8, 2],
    ...         'wings': [0, 2],
    ...     },
    ...     index=['spider', 'falcon'],
    ... )
    >>> other
            legs  wings
    spider     8      0
    falcon     2      2
    >>> df.filter_in(other)
            legs  wings
    falcon     2      2
    """

    return df[
        select_column(df, condition=condition)
        .isin(condition)
        .boolean(
            how=how,
            axis=1,
            complement=complement,
        )
    ]


def select_column(
    df: pd.DataFrame,
    /,
    condition: Iterable | SeriesOrFrame | dict[Hashable, list[Hashable]],
) -> pd.DataFrame:
    """Select DataFram columns via condition type"""

    if isinstance(condition, dict):
        # 'how' only works on condition these dictionary's keys
        return df[condition.keys()]
    elif isinstance(condition, pd.DataFrame):
        # 'how' only works on condition these DataFrame's columns
        return df[condition.columns]

    return df



from typing import Iterable

import pandas as pd

from dtoolkit.accessor.register import register_series_method


@register_series_method
def filter_in(
    s: pd.Series,
    condition: Iterable,
    /,
    complement: bool = False,
) -> pd.Series:
    """
    Filter :obj:`~pandas.Series` contents.

    Similar to :meth:`~pandas.Series.isin`, but the return is not bool.

    Parameters
    ----------
    condition : list-like
        The filtered result is based on this specific condition.

    complement : bool, default is False
        If True, do operation reversely.

    Returns
    -------
    Series

    See Also
    --------
    pandas.Series.isin
    pandas.Series.filter
    dtoolkit.accessor.dataframe.filter_in
        Filter DataFrame contents.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series(
    ...     ['lama', 'cow', 'lama', 'beetle', 'lama', 'hippo'],
    ...     name='animal',
    ... )
    >>> s
    0      lama
    1       cow
    2      lama
    3    beetle
    4      lama
    5     hippo
    Name: animal, dtype: object
    >>> s.filter_in(['cow', 'lama'])
    0      lama
    1       cow
    2      lama
    4      lama
    Name: animal, dtype: object
    >>> s.filter_in(['cow', 'lama'], complement=True)
    3    beetle
    5     hippo
    Name: animal, dtype: object
    """

    return s[invert_or_not(s.isin(condition), invert=complement)]


def invert_or_not(s: pd.Series, /, invert: bool = False) -> pd.Series:
    """
    Invert (~) the Series.

    Parameters
    ----------
    invert : bool, default is False
        If True, invert the Series.

    Returns
    -------
    Series

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series([True, False, True])
    >>> s
    0     True
    1    False
    2     True
    dtype: bool
    >>> s.pipe(invert_or_not)
    0     True
    1    False
    2     True
    dtype: bool
    >>> s.pipe(invert_or_not, invert=True)
    0    False
    1     True
    2    False
    dtype: bool
    """

    return ~s if invert else s



import os
import sys

from setuptools import setup

# Ensure the current directory is on sys.path so versioneer can be imported
# when pip uses PEP 517/518 build rules.
# https://github.com/python-versioneer/python-versioneer/issues/193
sys.path.append(os.path.dirname(__file__))

from versioneer import get_cmdclass  # noqa: E402
from versioneer import get_version  # noqa: E402


setup(
    version=get_version(),
    cmdclass=get_cmdclass(),
)



from __future__ import annotations

from typing import Literal

import numpy as np
import pandas as pd

from dtoolkit.accessor.register import register_series_method


@register_series_method
def drop_inf(
    s: pd.Series,
    /,
    inf: Literal["all", "pos", "+", "neg", "-"] = "all",
) -> pd.Series:
    """
    Remove ``inf`` values.

    Parameters
    ----------
    inf : {'all', 'pos', '+', 'neg', '-'}, default 'all'

        * 'all' : Remove ``inf`` and ``-inf``.
        * 'pos' / '+' : Only remove ``inf``.
        * 'neg' / '-' : Only remove ``-inf``.

    Returns
    -------
    Series
        Series with ``inf`` entries dropped from it.

    Raises
    ------
    ValueError
        If ``inf`` isn't "all", "pos", "+", "neg", or "-".

    See Also
    --------
    dtoolkit.accessor.dataframe.drop_inf
        :obj:`~pandas.DataFrame` drops rows or columns which contain ``inf``
        values.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> import numpy as np
    >>> s = pd.Series([1., 2., np.inf])
    >>> s
    0    1.0
    1    2.0
    2    inf
    dtype: float64

    Drop inf values from a Series.

    >>> s.drop_inf()
    0    1.0
    1    2.0
    dtype: float64
    """

    inf_range = get_inf_range(inf)
    mask = s.isin(inf_range)

    return s[~mask]


def get_inf_range(inf: Literal["all", "pos", "+", "neg", "-"] = "all") -> set[float]:
    """Get inf value from string"""

    INF_RANGE = {
        "all": {np.inf, -np.inf},
        "pos": {np.inf},
        "+": {np.inf},
        "neg": {-np.inf},
        "-": {-np.inf},
    }
    try:
        return INF_RANGE[inf]
    except KeyError as e:
        raise ValueError(f"invalid inf option: {inf!r}") from e



from __future__ import annotations

import numpy as np
import pandas as pd

from dtoolkit._typing import SeriesOrFrame
from dtoolkit._typing import TwoDimArray
from dtoolkit.accessor.dataframe import to_series  # noqa: F401


def transform_array_to_frame(
    array: np.ndarray,
    frame: SeriesOrFrame,
) -> np.ndarray | SeriesOrFrame:
    """
    Transform ``array``'s :obj:`type` (:obj:`~numpy.ndarray`) to
    :obj:`type` (:obj:`~pandas.Series` or :obj:`~pandas.DataFrame`).

    Parameters
    ----------
    array : array-like of shape ``(n_samples, n_features)``
    frame : Series or DataFrame

    Returns
    -------
    Series, DataFrame or ndarray
    """

    if (
        isinstance(array, (pd.Series, pd.DataFrame))
        or not isinstance(frame, (pd.Series, pd.DataFrame))
        or array.ndim > 2
    ):
        return array

    # Only length is equal
    # sparse matrix can't use `len`
    if array.shape[0] == frame.shape[0] and (
        array.ndim == 1 or (array.ndim == 2 and array.shape[1] == 1)
    ):
        if isinstance(frame, pd.Series):
            name = frame.name
        elif isinstance(frame, pd.DataFrame) and len(frame.columns) == 1:
            name = frame.columns[0]
        else:
            name = None

        return pd.Series(np.ravel(array), index=frame.index, name=name)

    # Both width and length are equal
    elif array.ndim == 2 and array.shape == frame.shape:
        return pd.DataFrame(array, columns=frame.columns, index=frame.index)

    return array


def transform_series_to_frame(X: np.ndarray | SeriesOrFrame) -> TwoDimArray:
    """
    Transform ``X`` to DataFrame if ``X`` type is Series.

    Parameters
    ----------
    X : ndarray, Series or DataFrame

    Returns
    -------
    DataFrame or ndarray
    """

    return X.to_frame() if isinstance(X, pd.Series) else X


def transform_frame_to_series(
    X: np.ndarray | SeriesOrFrame,
    drop_name: bool = False,
) -> np.ndarray | SeriesOrFrame:
    """
    Transform ``X`` to Series if ``X`` is one column DataFrame.

    Parameters
    ----------
    X : ndarray, Series or DataFrame
    drop_name : bool

    Returns
    -------
    ndarray, Series or DataFrame
    """

    if isinstance(X, pd.DataFrame):
        X = X.to_series()

    return X.rename() if drop_name and isinstance(X, pd.Series) else X


def snake_to_camel(name: str) -> str:
    """
    Change snake style name to camel style name.

    Parameters
    ----------
    name : str
        Snake style name

    Returns
    -------
    str
        Camel style name

    Examples
    --------
    >>> from dtoolkit.transformer._util import snake_to_camel
    >>> snake_to_camel(snake_to_camel.__name__)
    'SnakeToCamel'
    """

    components = name.split("_")
    components = (x.title() for x in components)
    return "".join(components)



from __future__ import annotations

import pandas as pd
from pandas.api.types import is_number

from dtoolkit.accessor.register import register_series_method


@register_series_method
def len(s: pd.Series, /, number: int = 1, other: int = None) -> pd.Series:
    """
    Return the length of each element in the series.

    Equals to ``s.apply(len)``, but the length of ``number`` type will as ``1``,
    the length of other types will as ``NaN``.

    Parameters
    ----------
    number : int, default 1
        The default length of `number` type.

    other : int or None, default None
        The default length of `other` type.

    Returns
    -------
    Series

    Notes
    -----
    - To keep the Python naming style, so use this accessor via
      ``Series.len`` rather than ``Series.lens``.

    - Different to :meth:`pandas.Series.str.len`. It only returns
      :class:`collections.abc.Iterable` type length. Other type will return `NaN`.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series([0, 1, 1.5, "string", ("tuple",), ["list"], {}, object])
    >>> s
    0                   0
    1                   1
    2                 1.5
    3              string
    4            (tuple,)
    5              [list]
    6                  {}
    7    <class 'object'>
    dtype: object
    >>> s.len()
    0    1.0
    1    1.0
    2    1.0
    3    6.0
    4    1.0
    5    1.0
    6    0.0
    7    NaN
    dtype: float64

    Set `number` and `other` default return.

    >>> s.len(number=0, other=0)
    0    0
    1    0
    2    0
    3    6
    4    1
    5    1
    6    0
    7    0
    dtype: int64
    """

    return s.apply(_wrap_len, number=number, other=other)


def _wrap_len(x, number: int, other: int | None) -> int | None:
    if hasattr(x, "__len__"):
        return x.__len__()
    elif is_number(x):
        return number

    return other



import pandas as pd

from dtoolkit.transformer.base import DataFrameTF


class GetTF(DataFrameTF):
    """
    A transformer could get item from object for given key
    (ex: :obj:`~pandas.DataFrame` column).

    See Also
    --------
    pandas.DataFrame.get : This transformer's prototype method.
    """

    transform_method = staticmethod(pd.DataFrame.get)



import pandas as pd

from dtoolkit.accessor.register import register_dataframe_method


@register_dataframe_method
def drop_or_not(df: pd.DataFrame, /, drop: bool = True, **kwargs) -> pd.DataFrame:
    """
    Drop specified labels from rows or columns.

    A sugary syntax wraps :meth:`~pandas.DataFrame.drop`::

        df.drop(**kwargs) if drop else df

    Parameters
    ----------
    drop : bool, default True
        Choose to drop or not. If True will drop else don't.

    **kwargs
        See the documentation for :meth:`~pandas.DataFrame.drop` for complete
        details on the keyword arguments.

    Returns
    -------
    DataFrame
        If ``drop=True`` will execute else return the original.

    See Also
    --------
    pandas.DataFrame.drop

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> df = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    >>> df
       a  b
    0  1  3
    1  2  4
    >>> df.drop_or_not(True, columns="b")
       a
    0  1
    1  2
    >>> df.drop_or_not(False, columns=["a"])
       a  b
    0  1  3
    1  2  4
    """

    return df.drop(**kwargs) if drop else df



import pandas as pd

from dtoolkit.accessor.register import register_series_method


@register_series_method("cut")
@register_series_method
def bin(s: pd.Series, /, *args, **kwargs) -> pd.Series:
    """
    Bin values into discrete intervals.

    A sugary syntax wraps :meth:`~pandas.cut`::

        pd.cut(s, *args, **kwargs)

    Parameters
    ----------
    *args, **kwargs
        See the documentation for :meth:`pandas.cut` for complete details on
        the positional arguments and the keyword arguments.

    Returns
    -------
    Series

    See Also
    --------
    pandas.cut
    dtoolkit.accessor.series.jenks_bin

    Notes
    -----
    This method could be called via ``s.bin`` or ``s.cut``.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd

    Create **score** samples:

    >>> s = pd.Series([100, 10, 50, 20, 90, 60])

    Bin score to rank level:

        - (0, 60] -> E
        - (60, 70] -> D
        - (70, 80] -> C
        - (80, 90] -> B
        - (90, 100] -> A

    >>> s.bin([0, 60, 70, 80, 90, 100], labels=['E', 'D', 'C', 'B', 'A'], right=True)
    0    A
    1    E
    2    E
    3    E
    4    B
    5    E
    dtype: category
    Categories (5, object): ['E' < 'D' < 'C' < 'B' < 'A']
    """

    return pd.cut(s, *args, **kwargs)



import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries import hole_counts as s_hole_counts
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_hole_counts, alias="df")
def hole_counts(df: gpd.GeoDataFrame, /) -> pd.Series:
    return s_hole_counts(df.geometry)



import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries import count_coordinates as s_count_coordinates
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_count_coordinates, klass="GeoDataFrame")
def count_coordinates(df: gpd.GeoDataFrame, /) -> pd.Series:
    return s_count_coordinates(df.geometry)



import geopandas as gpd
import pandas as pd

from dtoolkit.geoaccessor.geoseries.hole_counts import hole_counts
from dtoolkit.geoaccessor.register import register_geoseries_method


@register_geoseries_method
def has_hole(s: gpd.GeoSeries, /) -> pd.Series:
    """
    Check if each Polygon geometries have holes.

    Except for Polygon and MultiPolygon, other geometries will return ``False``.

    Returns
    -------
    Series

    See Also
    --------
    geopandas.GeoSeries.convex_hull : Drop holes in Polygon and MultiPolygon.
    dtoolkit.geoaccessor.geoseries.has_hole
    dtoolkit.geoaccessor.geoseries.hole_counts
    dtoolkit.geoaccessor.geodataframe.has_hole
    dtoolkit.geoaccessor.geodataframe.hole_counts

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import geopandas as gpd
    >>> from shapely.geometry import LineString, Point, Polygon
    >>> df = gpd.GeoDataFrame(
    ...     geometry=[
    ...         Polygon(
    ...             [(0, 0), (0, 5), (5, 5), (5, 0)],
    ...             [[(1, 1), (2, 1), (1, 2)], [(1, 4), (2, 4), (2, 3)]],
    ...         ),
    ...         Polygon([(1, 0), (2, 1), (0, 0)]),
    ...         LineString([(0, 0), (0, 5), (5, 5), (5, 0)]),
    ...         Point(0, 0),
    ...     ]
    ... )
    >>> df
                                                geometry
    0  POLYGON ((0.00000 0.00000, 0.00000 5.00000, 5....
    1  POLYGON ((1.00000 0.00000, 2.00000 1.00000, 0....
    2  LINESTRING (0.00000 0.00000, 0.00000 5.00000, ...
    3                            POINT (0.00000 0.00000)
    >>> df.has_hole()
    0     True
    1    False
    2    False
    3    False
    dtype: bool
    """

    return hole_counts(s) >= 1



import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.accessor.register import register_dataframe_method
from dtoolkit.accessor.series import set_unique_index as s_set_unique_index


@register_dataframe_method
@doc(s_set_unique_index, klass="DataFrame", alias="df")
def set_unique_index(df: pd.DataFrame, /, **kwargs) -> pd.DataFrame:
    return s_set_unique_index(df, **kwargs)



from __future__ import annotations

from warnings import warn

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry.base import BaseGeometry

from dtoolkit.geoaccessor.register import register_geoseries_method
from dtoolkit.util._exception import find_stack_level


@register_geoseries_method
def geodistance(
    s: gpd.GeoSeries,
    /,
    other: BaseGeometry | gpd.GeoSeries | gpd.GeoDataFrame,
    align: bool = True,
    radius: float = 6371008.7714150598,
) -> pd.Series:
    """
    Returns a ``Series`` containing the `great-circle`__ distance to aligned other
    via haversine formula.

    __ https://en.wikipedia.org/wiki/Great-circle_distance

    .. math::

        D(x, y) = 2 \\arcsin [
            \\sqrt{
                \\sin^2 ((y_2 - y_1) / 2)
                + \\cos(y_1) \\cos(y_1) \\sin^2 ((x_2 - y_1) / 2)
            }
        ]

    Parameters
    ----------
    other : BaseGeometry, GeoSeries, or GeoDataFrame

    align : bool, default True
        If True, automatically aligns GeoSeries based on their indices. If False,
        the order of elements is preserved.

    radius : float, default 6371008.7714150598
        Great-circle distance uses a spherical model of the earth, using the mean earth
        radius as defined by the International Union of Geodesy and Geophysics,
        (2\\ *a* + *b*)/3 = 6371008.7714150598 meters for WGS-84.

    Returns
    -------
    Series
        The values are the great-circle distances and its unit is meters.

    Raises
    ------
    ValueError
        If the CRS is not ``ESGP:4326``.

    TypeError
        If the other is not a ``BaseGeometry``, ``GeoSeries``, or ``GeoDataFrame``.

    See Also
    --------
    geopandas.GeoSeries.distance
    dtoolkit.geoaccessor.geoseries.geodistance
    dtoolkit.geoaccessor.geoseries.geodistance_matrix
    dtoolkit.geoaccessor.geodataframe.geodistance
    dtoolkit.geoaccessor.geodataframe.geodistance_matrix

    Notes
    -----
    - Currently, only supports Point geometry.
    - The great-circle distance is the angular distance between two points on the
      surface of a sphere. As the Earth is nearly spherical, the haversine formula
      provides a good approximation of the distance between two points of the Earth
      surface, with a less than 1% error on average.

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> import geopandas as gpd
    >>> from shapely.geometry import Point
    >>> df = (
    ...     pd.DataFrame({"x": [122, 100], "y":[55, 1]})
    ...     .from_xy("x", "y", crs=4326)
    ... )
    >>> df
         x   y                    geometry
    0  122  55  POINT (122.00000 55.00000)
    1  100   1   POINT (100.00000 1.00000)
    >>> df.geodistance(Point(120, 30)) / 1e6
    0    2.784435
    1    3.855604
    dtype: float64

    Calculate the great-circle distance of paired points.

    >>> s = gpd.GeoSeries([Point(120, 30), Point(120, 50)], index=[1, 2], crs=4326)
    >>> s
    1    POINT (120.00000 30.00000)
    2    POINT (120.00000 50.00000)
    dtype: geometry
    >>> df.geodistance(s) / 1e6
    0         NaN
    1    3.855604
    2         NaN
    dtype: float64
    >>> df.geodistance(s, align=False) / 1e6
    0    2.784435
    1    5.768885
    dtype: float64
    """

    if s.crs != 4326:
        raise ValueError(f"Only support 'EPSG:4326' CRS, but got {s.crs!r}.")

    if isinstance(other, BaseGeometry):
        x2, y2 = other.x, other.y
    elif isinstance(other, gpd.base.GeoPandasBase):
        if other.crs != 4326:
            raise ValueError(f"Only support 'EPSG:4326' CRS, but got {other.crs!r}.")
        if align and not s.index.equals(other.index):
            warn("The indices are different.", stacklevel=find_stack_level())
            s, other = s.align(other)

        x2, y2 = other.geometry.x.to_numpy(), other.geometry.y.to_numpy()
    else:
        raise TypeError(f"Unknown type: {type(other).__name__!r}.")

    x1, y1 = s.x.to_numpy(), s.y.to_numpy()
    return pd.Series(
        radius * haversine(x1, y1, x2, y2),
        index=s.index,
    )


def haversine(x1, y1, x2, y2) -> np.ndarray | float:
    """
    Compute the paired the great-circle distance between two points on the earth via
    haversine formula.

    .. math::

        D(x, y) = 2 \\arcsin [
            \\sqrt{
                \\sin^2 ((y_2 - y_1) / 2)
                + \\cos(y_1) \\cos(y_1) \\sin^2 ((x_2 - y_1) / 2)
            }
        ]

    Parameters
    ----------
    x1, y1 : array-like or float
        A feature array. The shape of data should be equal.

    x2, y2 : array-like or float
        A second feature array. The shape of data should be equal.

    Returns
    -------
    ndarray or float

    Notes
    -----
    The first coordinate of each point is assumed to be the longitude, the second is
    the latitude.

    Examples
    --------
    >>> bsas = [-58.5166646, -34.83333]
    >>> paris = [2.53844117956, 49.0083899664]

    Multiply by Earth radius to get kilometers

    >>> int(haversine(*bsas, *paris) * 63711)
    110997
    """

    x1, y1, x2, y2 = map(np.radians, (x1, y1, x2, y2))
    return 2 * np.arcsin(
        np.sqrt(
            np.sin((y2 - y1) / 2) ** 2
            + np.cos(y1) * np.cos(y2) * np.sin((x2 - x1) / 2) ** 2,
        ),
    )
